/*---------------------------------------------------------------------------*/
/* File:     EDIT_SCR.C                                                      */
/* Author:   R. P. Hanson                                                    */
/* Copyright:R. P. Hanson 1996                                               */
/*           Lothar Bittner is now the authorised custodian of the Atari ST  */
/*           version of this program.                                        */
/*---------------------------------------------------------------------------*/
/* Contents: Routines for updating voice data.                               */
/*---------------------------------------------------------------------------*/
/* Functions:"out_on_off"     display "on" or "off" on the display.          */
/*           "ch_val"         changes a value.                               */
/*           "ch_name"        changes a character in the voice name.         */
/*           "chngd"          checks a value to see if it has changed. If it */
/*                            has changed, then it displays the new value.   */
/*           "chngd_n"        Same as previous. Does not update display.     */
/*           "conv_num"       Converts a number to a fixed length string.    */
/*           "sel_vs_len"     Sets vector step repeat value.                 */
/*           "edit_all"       Routine that checks the permanent areas for    */
/*                            any changes.                                   */
/*           "chng_env"       Checks whether the envelope is being changed.  */
/*           "set_env"        Sets the envelope to one of the presets.       */
/*           "edit_element"   Checks for changes to element if one is in the */
/*                            main edit area.                                */
/*           "edit_global"    Checks for changes to voice common if it is in */
/*                            the main edit area.                            */
/*           "env_chngd"      Checks whether the envelope has changed.       */
/*           "upd_env"        Updates the envelope.                          */
/*           "upd_display"    Updates the main display.                      */
/*---------------------------------------------------------------------------*/
/* Globals:  "env_names"      array holding preset envelope names.           */
/*           "eff_names"      array holding full effect names.               */
/*           "sh_eff_names"   array holding short effect names.              */
/*           "step_length"    array holding vector step lengths.             */
/*           "vt_names"       array holding voice type prefixes.             */
/*           "env_store"      array holding preset envelopes.                */
/*           "str"            temporary store for outputting text.           */
/*           "first_time"     set to 1 when the screen is being redrawn.     */
/*           "chng_flag"      set when a parameter has changed.              */
/*           "last_MIDI_channel" holds the previously selected MIDI channel. */
/*           "synth_flag", "lt_synth_flag" indicate the current and previous */
/*                            selected synth (0=SY22, 1=SY35)                */
/*           "c_lvl_step", "c_dtn_step" holds the currently displayed step   */
/*                            in the level and detune vectors respectively.  */
/*           "lvl_chngd", "dtn_chngd" flags set when the level vector or     */
/*                            detune vector changes.                         */
/*           "mx", "my", "mbut", "kstat" holds the current mouse position,   */
/*                            mouse button and keyboard state.               */
/*           "upd_vect"       set to 1 when either vector changes.           */
/*           "voice_type", "last_voice_type" hold the current and previous   */
/*                            voice type (see "vt_names").                   */
/*---------------------------------------------------------------------------*/


/*----------*/
/* Includes */
/*----------*/
    #include <gemfast.h>
    #include <osbind.h>
    #include <string.h>
    #include "types.h"
    #include "sy_vdi.h"

/*--------------------*/
/* External variables */
/*--------------------*/
    extern unsigned char fm_data[256][21] ;
    extern unsigned char MIDI_channel ;
    extern char          fm_names[256][9] ;
    extern char          awm22_names[128][9] ;
    extern char          awm35_names[128][9] ;

/*----------------*/
/* Preset globals */
/*----------------*/
    char env_names[8][8] = {
        "...User", ".Preset", "..Piano", ".Guitar", "..Pluck", "..Brass",
        "Strings", "..Organ"
    } ;

    char eff_names[16][20] = {
        "Reverb Hall        ", "Reverb Room        ", "Reverb Plate       ",
        "Reverb Club        ", "Reverb Metal       ", "Single Short Delay ",
        "Long Delay 1       ", "Long Delay 2       ", "Doubler            ",
        "Ping-Pong Delay    ", "Panned Reflections ", "Early Reflections  ",
        "Gated Reverb       ", "Delay & Reverb 1   ", "Delay & Reverb 2   ",
        "Distortion & Reverb"
    } ;

    char sh_eff_names[16][9] = {
        "Rvb Hall", "Rvb Room", "RvbPlate", " RvbClub", "RvbMetal", " Delay 1",
        " Delay 2", " Delay 3", " Doubler", "PingPong", " Pan Ref", "EarlyRef",
        "GatedRev", "DlyRvb 1", "DlyRvb 2", "Dist Rvb"
    } ;

    char step_length[17][4] = {
        "160", " 10", " 20", " 30", " 40", " 50", " 60", " 70", " 80", " 90",
        "100", "110", "120", "130", "140", "150", "160"
    } ;

    char vt_names[18][3] = {
        "SP", "SC", "AP", "EP", "BR", "ST", "WN", "BA", "CH",
        "KY", "PL", "ME", "SE", "OR", "PC", "DR", "KS", "CT"
    } ;

    unsigned char env_store[6][8] = {
            { 87, 63, 127, 20, 119, 14,   0, 32 },
            {  0, 63, 127, 24, 117, 19,   0, 35 },
            { 87, 63, 127, 32,  79, 28,   0, 32 },
            {  0, 44, 127, 20, 119,  0, 119, 40 },
            { 87, 28, 127,  0, 127,  0, 127, 26 },
            {  0, 63, 127,  0, 127,  0, 127, 42 } } ;

/*------------------*/
/* Global variables */
/*------------------*/
    char          str[20] ;
    int           first_time ;
    int           chng_flag ;
    unsigned char last_MIDI_channel ;
    unsigned char synth_flag, lt_synth_flag ;
    unsigned char c_lvl_step, c_dtn_step ;
    unsigned char lvl_chngd, dtn_chngd ;
    int           mx, my, mbut, kstat ;
    int           upd_vect ;
    unsigned char voice_type, last_voice_type ;

/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "out_on_off"                                                */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function displays "Off" for a 0 otherwise "On".        */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "x", "y" : position to display at.                          */
/*              "v"      : value to display.                                */
/*--------------------------------------------------------------------------*/

    out_on_off ( x, y, v )
    int x, y ;
    unsigned char v ;
    {
        if ( v == 0 ) text_out ( x, y, "Off" ) ;
        else          text_out ( x, y, "On " ) ;
    }

/*---------------------------------*/
/*  END OF FUNCTION:  "out_on_off" */
/*---------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "ch_val"                                                    */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function checks to see if the mouse is within the      */
/*              defined region and if either button is pressed. If both     */
/*              conditions are true, then the input value is changed.       */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "y_bot"  : bottom of area to be checked.                    */
/*              "pol"    : button polarity. If 0, then left decreases value */
/*                         and right increases. Otherwise buttons swapped.  */
/*              "bottom" : lowest allowed value.                            */
/*              "top"    : highest allowed value.                           */
/*              "value"  : pointer to value to be changed.                  */
/*--------------------------------------------------------------------------*/

    ch_val ( y_bot, pol, bottom, top, value )
    int y_bot, pol ;
    int bottom, top ;
    char *value ;
    {
        int i_val, kstat ;

        /* Check that mouse is in y region */
        if (( my > y_bot-8 ) && ( my <= y_bot ))
        {
            i_val = *value ;
            if (( top > 127 ) && ( i_val < 0 ))  i_val &= 255 ;

            /* Get keybaord state */
            vq_key_s ( handle, &kstat ) ;

            /* Decrease value if required */
            if ( ( i_val > bottom ) && ((( mbut == 1 ) && ( pol == 0 )) ||
                                        (( mbut == 2 ) && ( pol == 1 ))))
            {
                if (( kstat & 3 ) == 0 )
                        i_val-- ;
                    else
                {
                    if ( ( i_val - bottom ) > 10 )
                        i_val -= 10 ;
                    else
                        i_val = bottom ;
                }
            }

            /* Increase value if required */
            if ( ( i_val < top ) && ((( mbut == 2 ) && ( pol == 0 )) ||
                                     (( mbut == 1 ) && ( pol == 1 ))))
            {
                if (( kstat & 3 ) == 0 )
                    i_val++ ;
                else
                {
                    if ( ( top - i_val ) > 10 )
                        i_val += 10 ;
                    else
                        i_val = top ;
                }
            }

            *value = i_val ;
        }
    }

/*-----------------------------*/
/*  END OF FUNCTION:  "ch_val" */
/*-----------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "ch_name"                                                   */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function checks to see if the mouse is within the      */
/*              defined region. If so then the character that the mouse is  */
/*              over is changed.                                            */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "x", "y" : mouse position.                                  */
/*--------------------------------------------------------------------------*/

    ch_name ( x, y )
    int x, y ;
    {
        int i, j ;

        for ( i = 0 ; i < 5 ; i++ )
        {
            j = (8 * i) + x ;
            if (( mx >= j ) && ( mx < (j + 7) ))
                ch_val ( y, 0, 32, 127, &curr_data.name[i+3] ) ;
        }
    }

/*------------------------------*/
/*  END OF FUNCTION:  "ch_name" */
/*------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "chngd"                                                     */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function checks to see if the input value has been     */
/*              changed. If x position is not 0, then it is updated on the  */
/*              display. If the value has changed, then global "chng_flag"  */
/*              is set to TRUE.                                             */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "x", "y" : output position.                                 */
/*              "new"    : pointer to value to be checked.                  */
/*--------------------------------------------------------------------------*/

    int chngd ( x, y, new )
    int x, y ;
    char *new ;
    {
        int ret_val, diff ;
        unsigned char *old_pt ;

        old_pt = (unsigned char *) &last_data ;
        diff = (long int) new - (long int) &curr_data ;

        ret_val = (( old_pt[diff] != *new ) || first_time ) ;
        if ( ret_val )
        {
            chng_flag = ret_val ;
            old_pt[diff] = *new ;
            sprintf ( str, "%d  ", *new ) ;
            if ( x != 0 )  text_out ( x, y, str ) ;
        }
        return ret_val ;
    }

/*----------------------------*/
/*  END OF FUNCTION:  "chngd" */
/*----------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "chngd_n"                                                   */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function checks to see if the input value has been     */
/*              changed. It calls "chgnd" with "x" and "y" set to 0 so that */
/*              the display is not updated. The reason this routine exists  */
/*              is due to a memory saving attempt when trying to compile    */
/*              the program with Megamax C.                                 */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "x", "y" : output position.                                 */
/*              "new"    : pointer to value to be checked.                  */
/*--------------------------------------------------------------------------*/

    int chngd_n ( new )
    char *new ;
    {
        return chngd ( 0, 0, new ) ;
    }

/*------------------------------*/
/*  END OF FUNCTION:  "chngd_n" */
/*------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "conv_num"                                                  */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function converts a number to a three digit string in  */
/*              global "str".                                               */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "i" : value to convert.                                     */
/*--------------------------------------------------------------------------*/

    conv_num ( i )
    int i ;
    {
        if ( i <= -10 )
            sprintf ( str, "%d", i ) ;
        else if (( i < 0 ) && ( i > -10 ))
            sprintf ( str, " %d", i ) ;
        else if (( i < 10 ) && ( i >= 0 ))
            sprintf ( str, "  %d", i ) ;
        else if ( i < 100 )
            sprintf ( str, " %d", i ) ;
        else
            sprintf ( str, "%d", i ) ;
    }

/*--------------------------------*/
/*  END OF FUNCTION:  "conv_num"  */
/*--------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "sel_vs_len"                                                */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function converts a vector step length to a value if   */
/*              less than 254, otherwise it selects "Rep" or "End".         */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "x", "y" : output position.                                 */
/*              "new"    : pointer to value to be checked.                  */
/*--------------------------------------------------------------------------*/

    char rep_end[2][4] = { "Rep", "End" } ;

    sel_vs_len ( i )
    unsigned char i ;
    {
        unsigned char j = i + 1 ;

        if ( i < 254 )
            conv_num ( j ) ;
        else
            strcpy ( str, rep_end[i-254] ) ;
    }

/*----------------------------------*/
/*  END OF FUNCTION:  "sel_vs_len"  */
/*----------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "edit_all"                                                  */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function checks whether anything is being changed in   */
/*              one of the permanent edit areas.                            */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: None.                                                       */
/*--------------------------------------------------------------------------*/

    edit_all ()
    {
        element_type *el_pt ;
        char loc_edit ;
        int y1, y2, y3, i ;

        /* Check area mouse is over according to y position */
        loc_edit = -1 ;
        if (( my > ( el_a_y1 - 7 )) && ( my <= el_a_y3 ))
            loc_edit = 0 ;
        else if (( my > ( el_b_y1 - 7 )) && ( my <= el_b_y3 ))
            loc_edit = 1 ;
        else if (( my > ( el_c_y1 - 7 )) && ( my <= el_c_y3 ))
            loc_edit = 2 ;
        else if (( my > ( el_d_y1 - 7 )) && ( my <= el_d_y3 ))
            loc_edit = 3 ;
        else if (( my > ( vect_y1 - 7 )) && ( my <= vect_y3 ))
            loc_edit = 4 ;
        else if (( my > ( glob_y1 - 7 )) && ( my <= glob_y3 ))
            loc_edit = 5 ;

        /* Check that the mouse is over one of these areas */
        if ( loc_edit != -1 )
        {
            if (( loc_edit >= 0 ) && ( loc_edit <= 3 ))
            {
                /* Element area... */

                /* Select local y values according to element */
                switch ( loc_edit )
                {
                    case 0 :    y1 = el_a_y1 ;  y2 = el_a_y2 ;  y3 = el_a_y3 ;
                                break ;
                    case 1 :    y1 = el_b_y1 ;  y2 = el_b_y2 ;  y3 = el_b_y3 ;
                                break ;
                    case 2 :    y1 = el_c_y1 ;  y2 = el_c_y2 ;  y3 = el_c_y3 ;
                                break ;
                    case 3 :    y1 = el_d_y1 ;  y2 = el_d_y2 ;  y3 = el_d_y3 ;
                                break ;
                }
                el_pt = &curr_data.element[loc_edit] ;

                /* Check velocity reponse */
                if (( mx >= 100 ) && ( mx <= 124 ))
                    ch_val ( y1, 0, -5, 5, &el_pt->vel_response ) ;

                /* Check element volume */
                if (( mx >= 10 ) && ( mx <= 124 ))
                {
                    ch_val ( y2, 1, 0, 127, &el_pt->osc_2.level ) ;
                    ch_val ( y3, 1, 0, 4,   &el_pt->pan ) ;
                }
            }
            else if ( loc_edit == 4 )
            {
                /* Vector area... */

                unsigned char l_last, v_last ;

                if (( mx >= 82 ) && ( mx <= 170 ))
                {
                    ch_val ( vect_y2, 0, 1, 16, &curr_data.v_level_rate ) ;
                    ch_val ( vect_y3, 0, 1, 16, &curr_data.v_detune_rate ) ;
                }
                if (( mx >= 200 ) && ( mx <= 256 ))
                {
                    l_last = c_lvl_step ; v_last = c_dtn_step ;
                    ch_val ( vect_y2, 0, 0, 49, &c_lvl_step ) ;
                    ch_val ( vect_y3, 0, 0, 49, &c_dtn_step ) ;
                }
                if (( mx >= 280 ) && ( mx < 336 ))
                {
                    ch_val ( vect_y2, 0, 0, 62, &curr_data.level_X[c_lvl_step] ) ;
                    ch_val ( vect_y3, 0, 0, 62, &curr_data.detune_X[c_dtn_step] ) ;
                }
                if (( mx >= 336 ) && ( mx < 392 ))
                {
                    ch_val ( vect_y2, 0, 0, 62, &curr_data.level_Y[c_lvl_step] ) ;
                    ch_val ( vect_y3, 0, 0, 62, &curr_data.detune_Y[c_dtn_step] ) ;
                }
                if (( mx > 392 ) && ( mx < 520 ))
                {
                    ch_val ( vect_y2, 0, 0, 255, &curr_data.level_step_length[c_lvl_step] ) ;
                    ch_val ( vect_y3, 0, 0, 255, &curr_data.detune_step_length[c_dtn_step] ) ;
                }

                for ( i = 0 ; i < 50 ; i++ )
                    if ( ( curr_data.level_X[c_lvl_step] !=
                                            last_data.level_X[c_lvl_step] ) ||
                         ( curr_data.level_Y[c_lvl_step] !=
                                            last_data.level_Y[c_lvl_step] ) ||
                         ( curr_data.detune_X[c_dtn_step] !=
                                            last_data.detune_X[c_dtn_step] ) ||
                         ( curr_data.detune_Y[c_dtn_step] !=
                                            last_data.detune_Y[c_dtn_step] ) ||
                         ( curr_data.level_step_length[c_lvl_step] !=
                                            last_data.level_step_length[c_lvl_step] ) ||
                         ( curr_data.detune_step_length[c_dtn_step] !=
                                            last_data.detune_step_length[c_dtn_step] ) )
                        upd_vect = 1 ;

                if ( chngd_n ( &curr_data.v_level_rate ) ||
                     ( l_last != c_lvl_step ) ||
                     chngd_n ( &curr_data.level_X[c_lvl_step] ) ||
                     chngd_n ( &curr_data.level_Y[c_lvl_step] ) ||
                     chngd_n ( &curr_data.level_step_length[c_lvl_step] ))
                    lvl_chngd = 1 ;

                if ( chngd_n ( &curr_data.v_detune_rate ) ||
                     chngd_n ( &curr_data.detune_X[c_dtn_step] ) ||
                     chngd_n ( &curr_data.detune_Y[c_dtn_step] ) ||
                     ( v_last != c_dtn_step ) ||
                     chngd_n ( &curr_data.detune_step_length[c_dtn_step] ))
                    dtn_chngd = 1 ;
            }
            else if ( loc_edit == 5 )
            {
                /* Voice common... */

                if (( mx >= 100 ) && ( mx < 124 ))
                    ch_val ( glob_y2, 0, 0, 17, &voice_type ) ;

                ch_name ( 124, glob_y2 ) ;

                if (( mx > 10 ) && ( mx < 180 ))
                    ch_val ( glob_y3, 0, 0, 1, &curr_data.num_elements ) ;

                if (( mx > 180 ) && ( mx < 326 ))
                {
                    ch_val ( glob_y1, 0, 0, 15, &curr_data.type ) ;
                    ch_val ( glob_y2, 0, 0, 7, &curr_data.effect ) ;
                    ch_val ( glob_y3, 0, 0, 12, &curr_data.pitch_bend ) ;
                }

                if (( mx > 326 ) && ( mx < 440 ))
                {
                    ch_val ( glob_y1, 0, 0, 1, &curr_data.mod_am_assign ) ;
                    ch_val ( glob_y2, 0, 0, 1, &curr_data.mod_pm_assign ) ;
                    ch_val ( glob_y3, 0, 0, 1, &curr_data.at_am_assign ) ;
                }

                if (( mx > 440 ) && ( mx < 554 ))
                {
                    ch_val ( glob_y1, 0, 0, 1, &curr_data.at_pm_assign ) ;
                    ch_val ( glob_y2, 0, 0, 1, &curr_data.at_lvl_assign ) ;
                    ch_val ( glob_y3, 0, -12, 12, &curr_data.at_bend ) ;
                }
            }
        }
    }

/*--------------------------------*/
/*  END OF FUNCTION:  "edit_all"  */
/*--------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "chng_env"                                                  */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function checks whether any parameter of an envelope   */
/*              is to be changed.                                           */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "env" : pointer to envelope structure to update.            */
/*--------------------------------------------------------------------------*/

    chng_env ( env )
    env_type *env ;
    {
        ch_val ( 272, 0, 0, 1,   &env->delay_switch ) ;
        ch_val ( 280, 1, 0, 127, &env->initial_level ) ;
        ch_val ( 288, 0, 0, 63,  &env->attack_rate ) ;
        ch_val ( 296, 1, 0, 127, &env->attack_level ) ;
        ch_val ( 304, 0, 0, 63,  &env->drate_1 ) ;
        ch_val ( 312, 1, 0, 127, &env->dlevel_1 ) ;
        ch_val ( 320, 0, 0, 63,  &env->drate_2 ) ;
        ch_val ( 328, 1, 0, 127, &env->dlevel_2 ) ;
        ch_val ( 336, 0, 0, 63,  &env->release_rate ) ;
    }

/*--------------------------------*/
/*  END OF FUNCTION:  "chng_env"  */
/*--------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "set_env"                                                   */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function sets the envelope to one of the presets.      */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "env" : pointer to envelope to update.                      */
/*              "i"   : preset to select.                                   */
/*--------------------------------------------------------------------------*/

    set_env ( env, i )
    env_type *env ;
    int i ;
    {
        env->initial_level = 127 - env_store[i][0] ;
        env->attack_rate   = env_store[i][1] ;
        env->attack_level  = 127 - env_store[i][2] ;
        env->drate_1       = env_store[i][3] ;
        env->dlevel_1      = 127 - env_store[i][4] ;
        env->drate_2       = env_store[i][5] ;
        env->dlevel_2      = 127 - env_store[i][6] ;
        env->release_rate  = env_store[i][7] ;
    }

/*-------------------------------*/
/*  END OF FUNCTION:  "set_env"  */
/*-------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "edit_element"                                              */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function is called when an element is in the edit area */
/*              at the bottom of the screen. It checks to see if any        */
/*              parameters are to be updated.                               */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: None.                                                       */
/*--------------------------------------------------------------------------*/

    edit_element ()
    {
        element_type *el_pt ;
        int i ;

        el_pt = &curr_data.element[curr_edit] ;

        if (( mx > 10 ) && ( mx < 218 ))
        {
            int wave_change ;

            /* Check waveform */
            if (( my > 224 ) && ( my <= 232 ))
            {
                wave_change = 0 ;
                if ( mbut == 2 )  { el_pt->wave++ ; wave_change = 1 ; }
                if ( mbut == 1 )  { el_pt->wave-- ; wave_change = 1 ; }
            }
            if (( curr_edit == 1 ) || ( curr_edit == 3 ))
                el_pt->wave &= 0xFF ;
            else
                el_pt->wave &= 0x7F ;

            /* If waveform changed and element is FM, select preset */
            if (( wave_change == 1 ) && (( curr_edit == 1 ) ||
                                         ( curr_edit == 3 )))
            {
                osc_type *osc_pt ;
                osc_pt = &el_pt->osc_2 ;
                osc_pt->fixed_frequency = fm_data[el_pt->wave][0] ;
                osc_pt->waveform = fm_data[el_pt->wave][1] ;
                osc_pt->frequency = fm_data[el_pt->wave][2] ;
                osc_pt = &el_pt->osc_1 ;
                osc_pt->level = fm_data[el_pt->wave][3] ;
                osc_pt->temperament = fm_data[el_pt->wave][4] ;
                osc_pt->detune = fm_data[el_pt->wave][5] ;
                osc_pt->level_scaling = fm_data[el_pt->wave][6] ;
                osc_pt->rate_scaling = fm_data[el_pt->wave][7] ;
                osc_pt->envelope_def.delay_switch = fm_data[el_pt->wave][8] ;
                osc_pt->envelope_def.attack_rate = fm_data[el_pt->wave][9] ;
                osc_pt->envelope_def.drate_1 = fm_data[el_pt->wave][10] ;
                osc_pt->envelope_def.drate_2 = fm_data[el_pt->wave][11] ;
                osc_pt->envelope_def.release_rate = fm_data[el_pt->wave][12] ;
                osc_pt->envelope_def.initial_level = fm_data[el_pt->wave][13] ;
                osc_pt->envelope_def.attack_level = fm_data[el_pt->wave][14] ;
                osc_pt->envelope_def.dlevel_1 = fm_data[el_pt->wave][15] ;
                osc_pt->envelope_def.dlevel_2 = fm_data[el_pt->wave][16] ;
                osc_pt->fixed_frequency = fm_data[el_pt->wave][17] ;
                osc_pt->waveform = fm_data[el_pt->wave][18] ;
                osc_pt->frequency = fm_data[el_pt->wave][19] ;
                el_pt->feedback = fm_data[el_pt->wave][20] ;
            }

            ch_val ( 240, 0, -5,   5, &el_pt->vel_response ) ;
            ch_val ( 248, 0, -3,   3, &el_pt->at_response ) ;
            ch_val ( 256, 0,  0,   4, &el_pt->LFO_wave ) ;
            ch_val ( 264, 0,  0,  31, &el_pt->LFO_speed ) ;
            ch_val ( 272, 0,  0, 255, &el_pt->LFO_delay ) ;
            ch_val ( 280, 0,  0, 127, &el_pt->LFO_rate ) ;
            ch_val ( 288, 0,  0,  15, &el_pt->LFO_AM_depth ) ;
            ch_val ( 296, 0,  0,  31, &el_pt->LFO_PM_depth ) ;

            /* If element is FM, check LFO assignments for updating */
            if (( curr_edit == 1 ) || ( curr_edit == 3 ))
            {
                ch_val ( 304, 0, 0, 1, &el_pt->AM_carrier ) ;
                ch_val ( 312, 0, 0, 1, &el_pt->AM_modulator ) ;
                ch_val ( 320, 0, 0, 1, &el_pt->PM_carrier ) ;
                ch_val ( 328, 0, 0, 1, &el_pt->PM_modulator ) ;
            }

            ch_val ( 336, 0, -12, 12, &el_pt->pitch_shift ) ;

            /* Check envelope type and select preset if changed */
            ch_val ( 344, 0, 0, 7, &el_pt->envelope ) ;
            if ( el_pt->envelope == 1 )
            {
                if ( mbut == 1 )  el_pt->envelope = 0 ;
                else              el_pt->envelope = 2 ;
            }

            if ( el_pt->envelope >= 2 )
                set_env ( &el_pt->osc_2.envelope_def, el_pt->envelope - 2 ) ;
        }

        /* Loop twice for FM element, once for AWM element */
        for ( i = 0 ; i < 2 ; i++ )
        {
            int x1, x2 ;
            osc_type *osc_ptr ;

            if (( curr_edit == 1 ) || ( curr_edit == 3 ))
            {
                if ( i == 0 ) { osc_ptr = &el_pt->osc_2 ; x1 = 210 ; }
                else          { osc_ptr = &el_pt->osc_1 ; x1 = 410 ; }
            }
            else
            {
                osc_ptr = &el_pt->osc_2 ;
                x1 = 210 ;
                i = 2 ;
            }

            x2 = x1 + 160 ;
            if (( mx > x1 ) && ( mx <= x2 ))
            {
                ch_val ( 232, 1, 0, 127, &osc_ptr->level ) ;
                ch_val ( 240, 0, 0,   3, &osc_ptr->temperament ) ;
                ch_val ( 248, 0, 0,  15, &osc_ptr->detune ) ;
                ch_val ( 256, 0, 0,  15, &osc_ptr->level_scaling ) ;
                ch_val ( 264, 0, 0,   7, &osc_ptr->rate_scaling ) ;

                chng_env ( &osc_ptr->envelope_def ) ;

                if (( curr_edit == 1 ) || ( curr_edit == 3 ))
                {
                    ch_val ( 344, 0, 0, 1, &osc_ptr->fixed_frequency ) ;
                    ch_val ( 352, 0, 0, 7, &osc_ptr->waveform ) ;
                    ch_val ( 360, 0, 0, 15, &osc_ptr->frequency ) ;

                    if ( i == 1 )
                        ch_val ( 368, 0, 0, 7, &el_pt->feedback ) ;
                }
            }
        }
    }

/*------------------------------------*/
/*  END OF FUNCTION:  "edit_element"  */
/*------------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "edit_global"                                               */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function is called when voice common is in the edit    */
/*              area at the bottom of the screen. It checks to see if any   */
/*              parameters are to be updated.                               */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: None.                                                       */
/*--------------------------------------------------------------------------*/

    edit_global ()
    {
        /* Check voice name for updating */
        if (( mx >= 178 ) && ( mx < 202 ))
            ch_val ( 232, 0, 0, 17, &voice_type ) ;
        ch_name ( 202, 232 ) ;

        if (( mx > 10 ) && ( mx < 338 ))
        {
            ch_val ( 240, 0,   0,  15, &curr_data.type ) ;
            ch_val ( 248, 0,   0,   7, &curr_data.effect ) ;
            ch_val ( 256, 0,   0,   1, &curr_data.num_elements ) ;
            ch_val ( 264, 0,   0,   1, &curr_data.mod_am_assign ) ;
            ch_val ( 272, 0,   0,   1, &curr_data.mod_pm_assign ) ;
            ch_val ( 280, 0,   0,   1, &curr_data.at_am_assign ) ;
            ch_val ( 288, 0,   0,   1, &curr_data.at_pm_assign ) ;
            ch_val ( 296, 0,   0,   1, &curr_data.at_lvl_assign ) ;
            ch_val ( 304, 0, -12,  12, &curr_data.at_bend ) ;
            ch_val ( 312, 0,   0, 127, &curr_data.env_delay ) ;
            ch_val ( 320, 0, -64,  63, &curr_data.attack_rate ) ;
            ch_val ( 328, 0, -64,  63, &curr_data.release_rate ) ;
            ch_val ( 336, 0,   0,  12, &curr_data.pitch_bend ) ;
            ch_val ( 344, 0,   0,  15, &MIDI_channel ) ;
            ch_val ( 352, 0,   0,   1, &synth_flag ) ;
        }
    }

/*-----------------------------------*/
/*  END OF FUNCTION:  "edit_global"  */
/*-----------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "env_changed"                                               */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function checks whether an envelope has changed.       */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "gc"           : value for result to be ORed with.          */
/*              "env1", "env2" : pointers to envelopes to be compared.      */
/*--------------------------------------------------------------------------*/

    int env_chngd ( gc, env1, env2 )
    int gc ;
    env_type *env1, *env2 ;
    {
        return ( gc || ( env1->delay_switch  != env2->delay_switch  ) ||
                       ( env1->initial_level != env2->initial_level ) ||
                       ( env1->attack_rate   != env2->attack_rate   ) ||
                       ( env1->attack_level  != env2->attack_level  ) ||
                       ( env1->drate_1       != env2->drate_1       ) ||
                       ( env1->dlevel_1      != env2->dlevel_1      ) ||
                       ( env1->drate_2       != env2->drate_2       ) ||
                       ( env1->dlevel_2      != env2->dlevel_2      ) ||
                       ( env1->release_rate  != env2->release_rate  )) ;
    }

/*---------------------------------*/
/*  END OF FUNCTION:  "env_chngd"  */
/*---------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "upd_env"                                                   */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function is called to check whether an envelope has    */
/*              been changed and needs to be re-displayed.                  */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "x"   : x position at which to write any updates.           */
/*              "env" : pointer to envelope to be checked.                  */
/*--------------------------------------------------------------------------*/

    upd_env ( x, env )
    int x ;
    env_type *env ;
    {
        unsigned char gp_temp ;

        if ( chngd_n ( &env->delay_switch ))
            out_on_off ( x, 272, env->delay_switch ) ;

        if ( chngd_n ( &env->initial_level ))
        {
            sprintf ( str, "%d ", 127 - env->initial_level ) ;
            text_out ( x, 280, str ) ;
        }

        /* Attack */
        chngd ( x, 288, &env->attack_rate ) ;
        if ( chngd_n ( &env->attack_level ))
        {
            sprintf ( str, "%d ", 127 - env->attack_level ) ;
            text_out ( x, 296, str ) ;
        }

        /* Decay 1 */
        chngd ( x, 304, &env->drate_1 ) ;
        if ( chngd_n ( &env->dlevel_1 ))
        {
            sprintf ( str, "%d ", 127 - env->dlevel_1 ) ;
            text_out ( x, 312, str ) ;
        }

        /* Decay 2 */
        chngd ( x, 320, &env->drate_2 ) ;
        if ( chngd_n ( &env->dlevel_2 ))
        {
            sprintf ( str, "%d ", 127 - env->dlevel_2 ) ;
            text_out ( x, 328, str ) ;
        }

        /* Release */
        chngd ( x, 336, &env->release_rate ) ;

        /* Calculate envelope level peak */
        env->env_peak = 0 ;
        gp_temp = env->initial_level ;
        if ( env->attack_level < gp_temp )
        {
            gp_temp = env->attack_level ;
            env->env_peak = 1 ;
        }
        if ( env->dlevel_1 < gp_temp )
        {
            gp_temp = env->dlevel_1 ;
            env->env_peak = 2 ;
        }
        if ( env->dlevel_2 < gp_temp )
            env->env_peak = 3 ;
    }

/*-------------------------------*/
/*  END OF FUNCTION:  "upd_env"  */
/*-------------------------------*/


/*--------------------------------------------------------------------------*/
/*  FUNCTION:   "upd_display"                                               */
/*--------------------------------------------------------------------------*/
/*  PURPOSE:    This function updates anything that has been changed.       */
/*--------------------------------------------------------------------------*/
/*  PARAMETERS: "first" : set to 1 if everything is to be updated.          */
/*--------------------------------------------------------------------------*/

    upd_display ( first )
    unsigned char first ;
    {
        element_type *el_pt, *el_pt_lt ;
        osc_type     *os_pt, *os_pt_lt ;
        int y1, y2, y3, i, j ;
        int glob_env_change ;
        unsigned char gp_temp ;

        glob_env_change = (( curr_data.attack_rate != last_data.attack_rate ) ||
                           ( curr_data.release_rate != last_data.release_rate ) ||
                           ( curr_data.env_delay != last_data.env_delay ) ||
                             first ) ;

        m_point_off ;

        first_time = first ;

        for ( i = 0 ; i < 4 ; i++ )
        {
            switch ( i )
            {
                case 0 :    y1 = el_a_y1 ;  y2 = el_a_y2 ;  y3 = el_a_y3 ;
                            break ;
                case 1 :    y1 = el_b_y1 ;  y2 = el_b_y2 ;  y3 = el_b_y3 ;
                            break ;
                case 2 :    y1 = el_c_y1 ;  y2 = el_c_y2 ;  y3 = el_c_y3 ;
                            break ;
                case 3 :    y1 = el_d_y1 ;  y2 = el_d_y2 ;  y3 = el_d_y3 ;
                            break ;
            }
            el_pt = &curr_data.element[i] ;
            el_pt_lt = &last_data.element[i] ;

            /* Update envelope graph if required */
            if ( env_chngd ( glob_env_change,
                               &el_pt->osc_1.envelope_def,
                               &el_pt_lt->osc_1.envelope_def ) ||
                 env_chngd ( glob_env_change,
                               &el_pt->osc_2.envelope_def,
                               &el_pt_lt->osc_2.envelope_def ))
            {
                draw_env (  y3, el_pt->osc_2.envelope_def, 0 ) ;
                if (( i == 1 ) || ( i == 3 ))
                    draw_env (  y3, el_pt->osc_1.envelope_def, 1 ) ;
            }

            if ( chngd_n ( &el_pt->vel_response ))
            {
                vel_out ( 100, y1, el_pt->vel_response, 24 ) ;
                if ( i == curr_edit )
                    vel_out ( 154, 241, el_pt->vel_response, 44 ) ;
            }

            if ( i == curr_edit )
            {
                if ( chngd_n ( &el_pt->wave ))
                {
                    if (( i == 1 ) || ( i == 3 ))
                        text_out ( 138, 232, fm_names[el_pt->wave] ) ;
                    else
                    {
                        if ( synth_flag == 0 )
                            text_out ( 138, 232, awm22_names[el_pt->wave] ) ;
                        else
                            text_out ( 138, 232, awm35_names[el_pt->wave] ) ;
                    }
                }

                if ( chngd_n ( &el_pt->at_response ))
                    AT_out ( el_pt->at_response ) ;

                if ( chngd_n ( &el_pt->LFO_wave ))
                    LFO_wv_out ( el_pt->LFO_wave ) ;

                chngd ( 154, 264, &el_pt->LFO_speed ) ;
                chngd ( 154, 272, &el_pt->LFO_delay ) ;
                chngd ( 154, 280, &el_pt->LFO_rate ) ;
                chngd ( 154, 288, &el_pt->LFO_AM_depth ) ;
                chngd ( 154, 296, &el_pt->LFO_PM_depth ) ;

                if (( i == 1 ) || ( i == 3 ))
                {
                    if ( chngd_n ( &el_pt->AM_carrier ))
                        out_on_off ( 154, 304, el_pt->AM_carrier ) ;

                    if ( chngd_n ( &el_pt->AM_modulator ))
                        out_on_off ( 154, 312, el_pt->AM_modulator ) ;

                    if ( chngd_n ( &el_pt->PM_carrier ))
                        out_on_off ( 154, 320, el_pt->PM_carrier ) ;

                    if ( chngd_n ( &el_pt->PM_modulator ))
                        out_on_off ( 154, 328, el_pt->PM_modulator ) ;
                }

                chngd ( 154, 336, &el_pt->pitch_shift ) ;

                if ( chngd_n ( &el_pt->envelope ))
                    text_out ( 130, 344, env_names[el_pt->envelope] ) ;
            }

            if ( chngd_n ( &el_pt->pan ))  pan_out ( y3, el_pt->pan ) ;

            for ( j = 0 ; j < 2 ; j++ )
            {
                int x ;

                if (( i == 1 ) || ( i == 3 ))
                {
                    if ( j == 0 )
                    {
                        os_pt = &el_pt->osc_2 ;
                        os_pt_lt = &el_pt_lt->osc_2 ;
                        x = 330 ;
                    }
                    else
                    {
                        os_pt = &el_pt->osc_1 ;
                        os_pt_lt = &el_pt_lt->osc_1 ;
                        x = 530 ;
                    }
                }
                else
                {
                    os_pt = &el_pt->osc_2 ;
                    os_pt_lt = &el_pt_lt->osc_2 ;
                    x = 330 ;
                    j = 2 ;
                }

                if ( chngd_n ( &os_pt->level ))
                {
                    if ( j != 1 )  level_out ( y2, 127 - os_pt->level ) ;
                    sprintf ( str, "%d  ", 127 - os_pt->level ) ;
                    if ( i == curr_edit )  text_out ( x, 232, str ) ;
                }

                if ( i == curr_edit )
                {
                    chngd ( x, 240, &os_pt->temperament ) ;
                    chngd ( x, 248, &os_pt->detune ) ;
                    if ( chngd_n ( &os_pt->level_scaling ))
                        ls_out ( x, os_pt->level_scaling ) ;

                    if ( chngd_n ( &os_pt->rate_scaling ))
                        rs_out ( x, os_pt->rate_scaling ) ;

                    upd_env ( x, &os_pt->envelope_def ) ;

                    if (( i == 1 ) || ( i == 3 ))
                    {
                        if ( chngd_n ( &os_pt->fixed_frequency ))
                            out_on_off ( x, 344, os_pt->fixed_frequency ) ;

                        chngd ( x, 352, &os_pt->waveform ) ;
                        chngd ( x, 360, &os_pt->frequency ) ;
                    }
                } /* END OF LOOP */

                if (( i == curr_edit ) && (( i == 1 ) || ( i == 3 )))
                    chngd ( 530, 368, &el_pt->feedback ) ;
            }
        }

        /* Check level vector */
        if (( lvl_chngd == 1 ) || first )
        {
            char ev ;
            chng_flag = ( lvl_chngd == 1 ) ;
            lvl_chngd = 0 ;
            sprintf ( str, "%d ", curr_data.v_level_rate * 10 ) ;
            text_out ( 138, vect_y2, str ) ;
            sprintf ( str, "%d ", c_lvl_step+1 ) ;
            text_out ( 240, vect_y2, str ) ;
            ev = curr_data.level_X[c_lvl_step] ;
            ev -= 31 ;
            conv_num ( ev ) ;
            text_out ( 304, vect_y2, str ) ;
            ev = curr_data.level_Y[c_lvl_step] ;
            ev -= 31 ;
            conv_num ( ev ) ;
            text_out ( 360, vect_y2, str ) ;
            sel_vs_len ( curr_data.level_step_length[c_lvl_step] ) ;
            text_out ( 440, vect_y2, str ) ;
        }

        /* Check detune vector */
        if (( dtn_chngd == 1 ) || first )
        {
            char ev ;
            chng_flag = ( dtn_chngd == 1 ) ;
            dtn_chngd = 0 ;
            sprintf ( str, "%d ", curr_data.v_detune_rate * 10 ) ;
            text_out ( 138, vect_y3, str ) ;
            sprintf ( str, "%d ", c_dtn_step+1 ) ;
            text_out ( 240, vect_y3, str ) ;
            ev = curr_data.detune_X[c_dtn_step] ;
            ev -= 31 ;
            conv_num ( ev ) ;
            text_out ( 304, vect_y3, str ) ;
            ev = curr_data.detune_Y[c_dtn_step] ;
            ev -= 31 ;
            conv_num ( ev ) ;
            text_out ( 360, vect_y3, str ) ;
            sel_vs_len ( curr_data.detune_step_length[c_dtn_step] ) ;
            text_out ( 440, vect_y3, str ) ;
        }

        /* Check voice type part of voice name */
        if ( voice_type != last_voice_type )
        {
            last_voice_type = voice_type ;
            curr_data.name[0] = vt_names[voice_type][0] ;
            curr_data.name[1] = vt_names[voice_type][1] ;
            curr_data.name[2] = '*' ;
        }

        /* Check rest of voice name */
        if (( strcmp ( last_data.name, curr_data.name ) != 0 ) || first )
        {
            chng_flag = TRUE ;
            strcpy ( last_data.name, curr_data.name ) ;
            text_out ( 100, glob_y2, curr_data.name ) ;
            if ( curr_edit == 5 )  text_out ( 178, 232, curr_data.name ) ;
        }

        if ( chngd_n ( &curr_data.type ))
        {
            text_out ( 252, glob_y1, sh_eff_names[curr_data.type] ) ;
            if ( curr_edit == 5 )
                text_out ( 178, 240, eff_names[curr_data.type] ) ;
        }

        if ( chngd ( 272, glob_y2, &curr_data.effect ))
            if ( curr_edit == 5 )  text_out ( 178, 248, str ) ;

        if ( chngd ( 272, glob_y3, &curr_data.pitch_bend ))
            if ( curr_edit == 5 )  text_out ( 178, 336, str ) ;

        if ( chngd_n ( &curr_data.num_elements ))
        {
            if ( curr_data.num_elements == 0 )
                strcpy ( str, "A-B    " ) ;
            else
                strcpy ( str, "A-B-C-D" ) ;
            text_out ( 84, glob_y3, str ) ;
            if ( curr_edit == 5 )  text_out ( 178, 256, str ) ;
        }

        if ( chngd_n ( &curr_data.mod_am_assign ))
        {
            out_on_off ( 400, glob_y1, curr_data.mod_am_assign ) ;
            if ( curr_edit == 5 )  out_on_off ( 178, 264, curr_data.mod_am_assign ) ;
        }

        if ( chngd_n ( &curr_data.mod_pm_assign ))
        {
            out_on_off ( 400, glob_y2, curr_data.mod_pm_assign ) ;
            if ( curr_edit == 5 )  out_on_off ( 178, 272, curr_data.mod_pm_assign ) ;
        }

        if ( chngd_n ( &curr_data.at_am_assign ))
        {
            out_on_off ( 400, glob_y3, curr_data.at_am_assign ) ;
            if ( curr_edit == 5 )  out_on_off ( 178, 280, curr_data.at_am_assign ) ;
        }

        if ( chngd_n ( &curr_data.at_pm_assign ))
        {
            out_on_off ( 532, glob_y1, curr_data.at_pm_assign ) ;
            if ( curr_edit == 5 )  out_on_off ( 178, 288, curr_data.at_pm_assign ) ;
        }

        if ( chngd_n ( &curr_data.at_lvl_assign ))
        {
            out_on_off ( 532, glob_y2, curr_data.at_lvl_assign ) ;
            if ( curr_edit == 5 )  out_on_off ( 178, 296, curr_data.at_lvl_assign ) ;
        }

        if ( chngd ( 532, glob_y3, &curr_data.at_bend ))
            if ( curr_edit == 5 )  text_out ( 178, 304, str ) ;

        if ( chngd_n ( &curr_data.env_delay ))
            if ( curr_edit == 5 )  text_out ( 178, 312, str ) ;

        if ( chngd_n ( &curr_data.attack_rate ))
            if ( curr_edit == 5 )  text_out ( 178, 320, str ) ;

        if ( chngd_n ( &curr_data.release_rate ))
            if ( curr_edit == 5 )  text_out ( 178, 328, str ) ;

        if ( curr_edit == 5 )
        {
            if ( chngd_n ( &synth_flag ))
            {
                if ( synth_flag == 0 )
                    strcpy ( str, "SY-22" ) ;
                else
                    strcpy ( str, "SY-35" ) ;
                text_out ( 178, 352, str ) ;
            }

            if (( last_MIDI_channel != MIDI_channel ) || first )
            {
                last_MIDI_channel = MIDI_channel ;
                chng_flag = TRUE ;
                sprintf ( str, "%d", MIDI_channel + 1 ) ;
                text_out ( 178, 344, str ) ;
            }
        }
        
        m_point_on ;
    }

/*-----------------------------------*/
/*  END OF FUNCTION:  "upd_display"  */
/*-----------------------------------*/
